"""
Sentry middleware for Inngest.

NOT STABLE! This is an experimental feature and may change in the future. If
you'd like to use it, we recommend copying this file into your source code.
"""

from __future__ import annotations

import sentry_sdk

import inngest


class SentryMiddleware(inngest.MiddlewareSync):
    """
    Middleware that adds Sentry tags and captures exceptions.
    """

    def __init__(
        self,
        client: inngest.Inngest,
        raw_request: object,
    ) -> None:
        """
        Args:
        ----
            client: Inngest client.
            raw_request: Framework/platform specific request object.
        """

        super().__init__(client, raw_request)

        if sentry_sdk.is_initialized() is False:
            client.logger.warning("Sentry SDK is not initialized")

        sentry_sdk.set_tag("inngest.app.id", client.app_id)

    def before_response(self) -> None:
        sentry_sdk.flush()

    def transform_input(
        self,
        ctx: inngest.Context,
        function: inngest.Function,
        steps: inngest.StepMemos,
    ) -> None:
        sentry_sdk.set_tag("inngest.event.count", len(ctx.events))
        sentry_sdk.set_tag("inngest.event.id", ctx.event.id)
        sentry_sdk.set_tag("inngest.event.name", ctx.event.name)
        sentry_sdk.set_tag("inngest.function.id", function.local_id)
        sentry_sdk.set_tag("inngest.function.name", function.name)
        sentry_sdk.set_tag("inngest.run.id", ctx.run_id)

    def transform_output(self, output: inngest.TransformOutputResult) -> None:
        if output.error:
            sentry_sdk.capture_exception(output.error)



from __future__ import annotations

import asyncio
import http
import os
import typing
import urllib.parse

import httpx
import typing_extensions

from inngest._internal import (
    client_lib,
    const,
    errors,
    execution_lib,
    function,
    middleware_lib,
    net,
    server_lib,
    step_lib,
    transforms,
    types,
)

from .models import (
    AuthenticatedInspection,
    CommResponse,
    UnauthenticatedInspection,
)
from .utils import parse_query_params

_ParamsT = typing_extensions.ParamSpec("_ParamsT")


def _prep_response(
    f: typing.Callable[
        _ParamsT, typing.Awaitable[typing.Union[CommResponse, Exception]]
    ],
) -> typing.Callable[_ParamsT, typing.Awaitable[CommResponse]]:
    async def inner(
        *args: _ParamsT.args,
        **kwargs: _ParamsT.kwargs,
    ) -> CommResponse:
        comm_handler = args[0]
        if not isinstance(comm_handler, CommHandler):
            raise ValueError("First argument must be a CommHandler instance.")

        res = await f(*args, **kwargs)
        if isinstance(res, Exception):
            res = CommResponse.from_error(comm_handler._client.logger, res)

        res.headers = {
            **res.headers,
            **net.create_headers(
                env=comm_handler._client.env,
                framework=comm_handler._framework,
                server_kind=comm_handler._client._mode,
            ),
        }

        return res

    return inner


def _prep_response_sync(
    f: typing.Callable[_ParamsT, typing.Union[CommResponse, Exception]],
) -> typing.Callable[_ParamsT, CommResponse]:
    def inner(
        *args: _ParamsT.args,
        **kwargs: _ParamsT.kwargs,
    ) -> CommResponse:
        comm_handler = args[0]
        if not isinstance(comm_handler, CommHandler):
            raise ValueError("First argument must be a CommHandler instance.")

        res = f(*args, **kwargs)
        if isinstance(res, Exception):
            res = CommResponse.from_error(comm_handler._client.logger, res)

        res.headers = {
            **res.headers,
            **net.create_headers(
                env=comm_handler._client.env,
                framework=comm_handler._framework,
                server_kind=comm_handler._client._mode,
            ),
        }

        return res

    return inner


class CommHandler:
    _base_url: str
    _client: client_lib.Inngest
    _fns: dict[str, function.Function]
    _framework: server_lib.Framework
    _mode: server_lib.ServerKind
    _signing_key: typing.Optional[str]
    _signing_key_fallback: typing.Optional[str]

    def __init__(
        self,
        *,
        client: client_lib.Inngest,
        framework: server_lib.Framework,
        functions: list[function.Function],
    ) -> None:
        self._client = client
        self._mode = client._mode
        self._api_origin = client.api_origin
        self._fns = {fn.get_id(): fn for fn in functions}
        self._framework = framework

        signing_key = client.signing_key
        if signing_key is None:
            if self._client.is_production:
                signing_key = os.getenv(const.EnvKey.SIGNING_KEY.value)
                if signing_key is None:
                    self._client.logger.error("missing signing key")
                    raise errors.SigningKeyMissingError()
        self._signing_key = signing_key

        self._signing_key_fallback = client.signing_key_fallback

    def _build_register_request(
        self,
        *,
        app_url: str,
        server_kind: typing.Optional[server_lib.ServerKind],
        sync_id: typing.Optional[str],
    ) -> types.MaybeError[httpx.Request]:
        registration_url = urllib.parse.urljoin(
            self._api_origin,
            "/fn/register",
        )

        fn_configs = self.get_function_configs(app_url)
        if isinstance(fn_configs, Exception):
            return fn_configs

        body = server_lib.SynchronizeRequest(
            app_name=self._client.app_id,
            deploy_type=server_lib.DeployType.PING,
            framework=self._framework,
            functions=fn_configs,
            sdk=f"{const.LANGUAGE}:v{const.VERSION}",
            url=app_url,
            v="0.1",
        ).to_dict()
        if isinstance(body, Exception):
            return body

        headers = net.create_headers(
            env=self._client.env,
            framework=self._framework,
            server_kind=server_kind,
        )

        params = {}
        if sync_id is not None:
            params[server_lib.QueryParamKey.SYNC_ID.value] = sync_id

        return self._client._http_client_sync.build_request(
            "POST",
            registration_url,
            headers=headers,
            json=transforms.deep_strip_none(body),
            params=params,
            timeout=30,
        )

    @_prep_response
    async def call_function(
        self,
        *,
        body: bytes,
        headers: typing.Union[dict[str, str], dict[str, list[str]]],
        query_params: typing.Union[dict[str, str], dict[str, list[str]]],
        raw_request: object,
    ) -> typing.Union[CommResponse, Exception]:
        """Handle a function call from the Executor."""

        headers = net.normalize_headers(headers)

        server_kind = transforms.get_server_kind(headers)
        if isinstance(server_kind, Exception):
            self._client.logger.error(server_kind)
            server_kind = None

        middleware = middleware_lib.MiddlewareManager.from_client(
            self._client,
            raw_request,
        )

        # Validate the request signature.
        err = net.validate_request(
            body=body,
            headers=headers,
            mode=self._client._mode,
            signing_key=self._signing_key,
            signing_key_fallback=self._signing_key_fallback,
        )
        if isinstance(err, Exception):
            return err

        request = server_lib.ServerRequest.from_raw(body)
        if isinstance(request, Exception):
            return request

        params = parse_query_params(query_params)
        if isinstance(params, Exception):
            return params
        if params.fn_id is None:
            return errors.QueryParamMissingError(
                server_lib.QueryParamKey.FUNCTION_ID.value
            )

        # Get the function we should call.
        fn = self._get_function(params.fn_id)
        if isinstance(fn, Exception):
            return fn

        events = request.events
        steps = request.steps
        if request.use_api:
            # Putting the batch and memoized steps in the request would make it
            # to big, so the Executor is telling the SDK to fetch them from the
            # API

            fetched_events, fetched_steps = await asyncio.gather(
                self._client._get_batch(request.ctx.run_id),
                self._client._get_steps(request.ctx.run_id),
            )
            if isinstance(fetched_events, Exception):
                return fetched_events
            events = fetched_events
            if isinstance(fetched_steps, Exception):
                return fetched_steps
            steps = fetched_steps
        if events is None:
            # Should be unreachable. The Executor should always either send the
            # batch or tell the SDK to fetch the batch

            return Exception("events not in request")
        call_res = await fn.call(
            self._client,
            execution_lib.Context(
                attempt=request.ctx.attempt,
                event=request.event,
                events=events,
                logger=self._client.logger,
                run_id=request.ctx.run_id,
            ),
            params.fn_id,
            middleware,
            request,
            step_lib.StepMemos.from_raw(steps),
            params.step_id,
        )

        return CommResponse.from_call_result(
            self._client.logger,
            call_res,
            self._client.env,
            self._framework,
            server_kind,
        )

    @_prep_response_sync
    def call_function_sync(
        self,
        *,
        body: bytes,
        headers: typing.Union[dict[str, str], dict[str, list[str]]],
        query_params: typing.Union[dict[str, str], dict[str, list[str]]],
        raw_request: object,
    ) -> typing.Union[CommResponse, Exception]:
        """Handle a function call from the Executor."""

        headers = net.normalize_headers(headers)

        server_kind = transforms.get_server_kind(headers)
        if isinstance(server_kind, Exception):
            self._client.logger.error(server_kind)
            server_kind = None

        middleware = middleware_lib.MiddlewareManager.from_client(
            self._client,
            raw_request,
        )

        # Validate the request signature.
        err = net.validate_request(
            body=body,
            headers=headers,
            mode=self._client._mode,
            signing_key=self._signing_key,
            signing_key_fallback=self._signing_key_fallback,
        )
        if isinstance(err, Exception):
            return err

        request = server_lib.ServerRequest.from_raw(body)
        if isinstance(request, Exception):
            return request

        params = parse_query_params(query_params)
        if isinstance(params, Exception):
            return params
        if params.fn_id is None:
            return errors.QueryParamMissingError(
                server_lib.QueryParamKey.FUNCTION_ID.value
            )

        # Get the function we should call.
        fn = self._get_function(params.fn_id)
        if isinstance(fn, Exception):
            return fn

        events = request.events
        steps = request.steps
        if request.use_api:
            # Putting the batch and memoized steps in the request would make it
            # to big, so the Executor is telling the SDK to fetch them from the
            # API

            fetched_events = self._client._get_batch_sync(request.ctx.run_id)
            if isinstance(fetched_events, Exception):
                return fetched_events
            events = fetched_events

            fetched_steps = self._client._get_steps_sync(request.ctx.run_id)
            if isinstance(fetched_steps, Exception):
                return fetched_steps
            steps = fetched_steps
        if events is None:
            # Should be unreachable. The Executor should always either send the
            # batch or tell the SDK to fetch the batch

            return Exception("events not in request")

        call_res = fn.call_sync(
            self._client,
            execution_lib.Context(
                attempt=request.ctx.attempt,
                event=request.event,
                events=events,
                logger=self._client.logger,
                run_id=request.ctx.run_id,
            ),
            params.fn_id,
            middleware,
            request,
            step_lib.StepMemos.from_raw(steps),
            params.step_id,
        )

        return CommResponse.from_call_result(
            self._client.logger,
            call_res,
            self._client.env,
            self._framework,
            server_kind,
        )

    def _get_function(self, fn_id: str) -> types.MaybeError[function.Function]:
        # Look for the function ID in the list of user functions, but also
        # look for it in the list of on_failure functions.
        for _fn in self._fns.values():
            if _fn.get_id() == fn_id:
                return _fn
            if _fn.on_failure_fn_id == fn_id:
                return _fn

        # If we didn't find the function ID, it might be because the function ID
        # in the request uses the old format that didn't include the app ID.
        # We'll prefix the function ID with the app ID and try again. This logic
        # can be deleted when no one is using Python SDK versions below 0.3.0
        # anymore.
        app_and_fn_id = f"{self._client.app_id}-{fn_id}"
        for _fn in self._fns.values():
            if _fn.get_id() == app_and_fn_id:
                return _fn
            if _fn.on_failure_fn_id == app_and_fn_id:
                return _fn

        return errors.FunctionNotFoundError(f"function {fn_id} not found")

    def get_function_configs(
        self,
        app_url: str,
    ) -> types.MaybeError[list[server_lib.FunctionConfig]]:
        configs: list[server_lib.FunctionConfig] = []
        for fn in self._fns.values():
            config = fn.get_config(app_url)
            configs.append(config.main)

            if config.on_failure is not None:
                configs.append(config.on_failure)

        if len(configs) == 0:
            return errors.FunctionConfigInvalidError("no functions found")
        return configs

    @_prep_response_sync
    def inspect(
        self,
        *,
        body: bytes,
        headers: typing.Union[dict[str, str], dict[str, list[str]]],
        serve_origin: typing.Optional[str],
        serve_path: typing.Optional[str],
    ) -> CommResponse:
        """Handle Dev Server's auto-discovery."""

        headers = net.normalize_headers(headers)

        server_kind = transforms.get_server_kind(headers)
        if isinstance(server_kind, Exception):
            self._client.logger.error(server_kind)
            server_kind = None

        if server_kind is not None and server_kind != self._mode:
            # Tell Dev Server to leave the app alone since it's in production
            # mode.
            return CommResponse(
                body={},
                status_code=403,
            )
        # Validate the request signature.
        err = net.validate_request(
            body=body,
            headers=headers,
            mode=self._client._mode,
            signing_key=self._signing_key,
            signing_key_fallback=self._signing_key_fallback,
        )
        if self._client._mode != server_lib.ServerKind.CLOUD or isinstance(
            err, Exception
        ):
            authentication_succeeded = None
            if isinstance(err, Exception):
                authentication_succeeded = False

            res_body = UnauthenticatedInspection(
                authentication_succeeded=authentication_succeeded,
                function_count=len(self._fns),
                has_event_key=self._client.event_key is not None,
                has_signing_key=self._signing_key is not None,
                has_signing_key_fallback=self._signing_key_fallback is not None,
                mode=self._mode,
            )
        else:
            event_key_hash = (
                transforms.hash_event_key(self._client.event_key)
                if self._client.event_key
                else None
            )

            signing_key_hash = (
                transforms.hash_signing_key(self._signing_key)
                if self._signing_key
                else None
            )

            signing_key_fallback_hash = (
                transforms.hash_signing_key(self._signing_key_fallback)
                if self._signing_key_fallback
                else None
            )

            res_body = AuthenticatedInspection(
                api_origin=self._client.api_origin,
                app_id=self._client.app_id,
                authentication_succeeded=True,
                env=self._client.env,
                event_api_origin=self._client.event_api_origin,
                event_key_hash=event_key_hash,
                framework=self._framework.value,
                function_count=len(self._fns),
                has_event_key=self._client.event_key is not None,
                has_signing_key=self._signing_key is not None,
                has_signing_key_fallback=self._signing_key_fallback is not None,
                mode=self._mode,
                serve_origin=serve_origin,
                serve_path=serve_path,
                signing_key_fallback_hash=signing_key_fallback_hash,
                signing_key_hash=signing_key_hash,
            )

        body_json = res_body.to_dict()
        if isinstance(body, Exception):
            body_json = {
                "error": "failed to serialize inspection data",
            }

        return CommResponse(
            body=body_json,
            status_code=200,
        )

    def _parse_registration_response(
        self,
        server_res: httpx.Response,
    ) -> CommResponse:
        try:
            server_res_body = server_res.json()
        except Exception:
            return CommResponse.from_error(
                self._client.logger,
                errors.RegistrationFailedError("response is not valid JSON"),
            )

        if not isinstance(server_res_body, dict):
            return CommResponse.from_error(
                self._client.logger,
                errors.RegistrationFailedError("response is not an object"),
            )

        if server_res.status_code < 400:
            return CommResponse(
                body=server_res_body,
                status_code=http.HTTPStatus.OK,
            )

        msg = server_res_body.get("error")
        if not isinstance(msg, str):
            msg = "registration failed"
        comm_res = CommResponse.from_error(
            self._client.logger,
            errors.RegistrationFailedError(msg.strip()),
        )
        comm_res.status_code = server_res.status_code
        return comm_res

    @_prep_response
    async def register(
        self: CommHandler,
        *,
        headers: typing.Union[dict[str, str], dict[str, list[str]]],
        query_params: typing.Union[dict[str, str], dict[str, list[str]]],
        request_url: str,
        serve_origin: typing.Optional[str],
        serve_path: typing.Optional[str],
    ) -> typing.Union[CommResponse, Exception]:
        """Handle a registration call."""

        headers = net.normalize_headers(headers)

        app_url = net.create_serve_url(
            request_url=request_url,
            serve_origin=serve_origin,
            serve_path=serve_path,
        )

        server_kind = transforms.get_server_kind(headers)
        if isinstance(server_kind, Exception):
            self._client.logger.error(server_kind)
            server_kind = None

        comm_res = self._validate_registration(server_kind)
        if comm_res is not None:
            return comm_res

        params = parse_query_params(query_params)
        if isinstance(params, Exception):
            return params

        req = self._build_register_request(
            app_url=app_url,
            server_kind=server_kind,
            sync_id=params.sync_id,
        )
        if isinstance(req, Exception):
            return req

        res = await net.fetch_with_auth_fallback(
            self._client._http_client,
            self._client._http_client_sync,
            req,
            signing_key=self._signing_key,
            signing_key_fallback=self._signing_key_fallback,
        )
        if isinstance(res, Exception):
            return res

        return self._parse_registration_response(res)

    @_prep_response_sync
    def register_sync(
        self: CommHandler,
        *,
        headers: typing.Union[dict[str, str], dict[str, list[str]]],
        query_params: typing.Union[dict[str, str], dict[str, list[str]]],
        request_url: str,
        serve_origin: typing.Optional[str],
        serve_path: typing.Optional[str],
    ) -> typing.Union[CommResponse, Exception]:
        """Handle a registration call."""

        headers = net.normalize_headers(headers)

        app_url = net.create_serve_url(
            request_url=request_url,
            serve_origin=serve_origin,
            serve_path=serve_path,
        )

        server_kind = transforms.get_server_kind(headers)
        if isinstance(server_kind, Exception):
            self._client.logger.error(server_kind)
            server_kind = None

        comm_res = self._validate_registration(server_kind)
        if comm_res is not None:
            return comm_res

        params = parse_query_params(query_params)
        if isinstance(params, Exception):
            return params

        req = self._build_register_request(
            app_url=app_url,
            server_kind=server_kind,
            sync_id=params.sync_id,
        )
        if isinstance(req, Exception):
            return req

        res = net.fetch_with_auth_fallback_sync(
            self._client._http_client_sync,
            req,
            signing_key=self._signing_key,
            signing_key_fallback=self._signing_key_fallback,
        )
        if isinstance(res, Exception):
            return res

        return self._parse_registration_response(res)

    def _validate_registration(
        self,
        server_kind: typing.Optional[server_lib.ServerKind],
    ) -> typing.Optional[CommResponse]:
        if server_kind is not None and server_kind != self._mode:
            msg: str
            if server_kind == server_lib.ServerKind.DEV_SERVER:
                msg = "Sync rejected since it's from a Dev Server but expected Cloud"
            else:
                msg = "Sync rejected since it's from Cloud but expected Dev Server"

            self._client.logger.error(msg)
            return CommResponse.from_error_code(
                server_lib.ErrorCode.SERVER_KIND_MISMATCH,
                msg,
                http.HTTPStatus.BAD_REQUEST,
            )

        return None



import os
import typing

from . import const, net, server_lib


def get_environment_name() -> typing.Optional[str]:
    """
    Get the Inngest environment name from env vars. Checks our env var and cloud
    platform (e.g. Vercel) specific env vars
    """

    return (
        os.getenv(const.EnvKey.ENV.value)
        or os.getenv(const.EnvKey.RAILWAY_GIT_BRANCH.value)
        or os.getenv(const.EnvKey.RENDER_GIT_BRANCH.value)
        or os.getenv(const.EnvKey.VERCEL_GIT_BRANCH.value)
    )


def get_url(
    env_var: const.EnvKey,
    mode: server_lib.ServerKind,
) -> typing.Optional[str]:
    """
    Get a URL from an env var. Returns None if the env var is not set or if its value is not a valid URL
    """

    val = os.getenv(env_var.value)
    if val is None:
        return None
    val = val.strip()

    parsed = net.parse_url(val, mode)
    if isinstance(parsed, Exception):
        return None

    return parsed


def is_truthy(env_var: const.EnvKey) -> bool:
    val = os.getenv(env_var.value)
    if val is None:
        return False
    val = val.strip()

    if val.lower() in ("false", "0", ""):
        return False

    return True



from __future__ import annotations

import typing

from inngest._internal import server_lib, types


class _QueryParams(types.BaseModel):
    fn_id: typing.Optional[str]
    step_id: typing.Optional[str]
    sync_id: typing.Optional[str]


def parse_query_params(
    query_params: typing.Union[dict[str, str], dict[str, list[str]]],
) -> typing.Union[_QueryParams, Exception]:
    normalized: dict[str, str] = {}
    for k, v in query_params.items():
        if isinstance(v, list):
            normalized[k] = v[0]
        else:
            normalized[k] = v

    step_id = normalized.get(server_lib.QueryParamKey.STEP_ID.value)
    if step_id == server_lib.UNSPECIFIED_STEP_ID:
        step_id = None

    return _QueryParams(
        fn_id=normalized.get(server_lib.QueryParamKey.FUNCTION_ID.value),
        step_id=step_id,
        sync_id=normalized.get(server_lib.QueryParamKey.SYNC_ID.value),
    )



"""
Encryption middleware for Inngest.

NOT STABLE! This is an experimental feature and may change in the future. If
you'd like to use it, we recommend copying this file into your source code.
"""

from __future__ import annotations

import json
import typing

import nacl.encoding
import nacl.secret
import nacl.utils

import inngest

encryption_marker: typing.Final = "__ENCRYPTED__"


class EncryptionMiddleware(inngest.MiddlewareSync):
    """
    Middleware that encrypts and decrypts data using a symmetric key. The
    cryptography library is a dependency.
    """

    def __init__(
        self,
        client: inngest.Inngest,
        raw_request: object,
        secret_key: typing.Union[bytes, str],
    ) -> None:
        """
        Args:
        ----
            client: Inngest client.
            raw_request: Framework/platform specific request object.
            secret_key: Fernet secret key used for encryption and decryption.
        """

        super().__init__(client, raw_request)

        if isinstance(secret_key, str):
            secret_key = bytes.fromhex(secret_key)

        self._box = nacl.secret.SecretBox(secret_key)

    @classmethod
    def factory(
        cls,
        secret_key: typing.Union[bytes, str],
    ) -> typing.Callable[[inngest.Inngest, object], EncryptionMiddleware]:
        """
        Create an encryption middleware factory that can be passed to an Inngest
        client or function.

        Args:
        ----
            secret_key: Fernet secret key used for encryption and decryption.
        """

        def _factory(
            client: inngest.Inngest,
            raw_request: object,
        ) -> EncryptionMiddleware:
            return cls(client, raw_request, secret_key)

        return _factory

    def _encrypt(self, data: object) -> dict[str, typing.Union[bool, str]]:
        if isinstance(data, dict) and data.get(encryption_marker) is True:
            # Already encrypted
            return data

        byt = json.dumps(data).encode()
        ciphertext = self._box.encrypt(
            byt,
            encoder=nacl.encoding.Base64Encoder,
        )
        return {
            encryption_marker: True,
            "data": ciphertext.decode(),
        }

    def _decrypt(self, data: object) -> inngest.JSON:
        if not _is_encrypted(data) or not isinstance(data, dict):
            # Not encrypted
            return data  # type: ignore

        encrypted = data.get("data")
        if not isinstance(encrypted, str):
            return data

        byt = self._box.decrypt(
            encrypted.encode(),
            encoder=nacl.encoding.Base64Encoder,
        )
        return json.loads(byt.decode())  # type: ignore

    def _decrypt_event_data(
        self,
        data: typing.Mapping[str, inngest.JSON],
    ) -> typing.Mapping[str, inngest.JSON]:
        is_everything_encrypted = _is_encrypted(data)
        if is_everything_encrypted:
            decrypted = self._decrypt(data)
            if isinstance(decrypted, dict):
                data = decrypted
        else:
            data = {k: self._decrypt(v) for k, v in data.items()}

        return data

    def before_send_events(self, events: list[inngest.Event]) -> None:
        """
        Encrypt event data before sending it to the Inngest server.
        """

        for event in events:
            event.data = self._encrypt(event.data)

    def transform_input(
        self,
        ctx: inngest.Context,
        function: inngest.Function,
        steps: inngest.StepMemos,
    ) -> None:
        """
        Decrypt data from the Inngest server.
        """

        for step in steps.values():
            step.data = self._decrypt(step.data)

        ctx.event.data = self._decrypt_event_data(ctx.event.data)

        for event in ctx.events:
            event.data = self._decrypt_event_data(event.data)

    def transform_output(self, result: inngest.TransformOutputResult) -> None:
        """
        Encrypt data before sending it to the Inngest server.
        """

        if result.output is not None:
            result.output = self._encrypt(result.output)


def _is_encrypted(value: object) -> bool:
    return isinstance(value, dict) and value.get(encryption_marker) is True
